module ParallelCoordinates exposing (..)

import Axis exposing (..)
import Color
import ExtendedFunctions exposing (listmap10, map10)
import Html exposing (Html, b, br, option, select, text)
import Html.Attributes exposing (align, selected, value)
import Html.Events exposing (onClick, onInput)
import List
import Maybe
import Model exposing (Model)
import Path
import Scale exposing (ContinuousScale)
import Set
import Shape
import SmartPhoneType exposing (Smartphone)
import Statistics
import TypedSvg exposing (g, style, svg, text_)
import TypedSvg.Attributes exposing (class, fill, fontSize, stroke, strokeWidth, textAnchor, transform, viewBox, y1, y2)
import TypedSvg.Attributes.InPx exposing (x, y)
import TypedSvg.Core exposing (Svg)
import TypedSvg.Types exposing (AnchorAlignment(..), Length(..), Paint(..), Transform(..))


w : Float
w =
    900


h : Float
h =
    430


padding : Float
padding =
    60


axisPaddig : Float
axisPaddig =
    40


tickCount : Int
tickCount =
    3


defaultExtent : ( number, number1 )
defaultExtent =
    ( 0, 200000 )


type Msg
    = ChangeOrder Int Int
    | SetScatterPlotFromParallelPlot String String
    | SetBrandParallel String
    | Set5GParallel String
    | SetNFCParallel String
    | SetIRBlasterParallel String


parallelCoordinatesPlot : Model -> List Smartphone -> Svg Msg
parallelCoordinatesPlot model smartphonesList =
    let
        getAttributeList : (Smartphone -> Maybe Float) -> List Smartphone -> List (Maybe Float)
        getAttributeList attribute smartphones =
            List.map attribute smartphones

        convertMaybeList : List (Maybe Float) -> Float -> List Float
        convertMaybeList maybeList defaultValue =
            List.map (Maybe.withDefault defaultValue) maybeList

        getAllAttributeLists : List (Smartphone -> Maybe Float) -> List Smartphone -> List (List Float)
        getAllAttributeLists attributes smartphones =
            List.map (\attribute -> convertMaybeList (getAttributeList attribute smartphones) 0.0) attributes

        multiDimValues =
            getAllAttributeLists model.parallelPlotOption.orderedList smartphonesList

        multidimStrings =
            model.parallelPlotOption.orderedLabels

        dimLen =
            List.length multiDimValues - 1

        yValues =
            multiDimValues

        yScaleLocals =
            List.map yScale yValues

        yScaleLocalsWithIndex =
            List.indexedMap Tuple.pair (List.map yScale yValues)

        xScaleLocal =
            xScale (List.map toFloat (List.range 0 dimLen))

        yaxis =
            List.map3 (createYAxis xScaleLocal model) yScaleLocalsWithIndex multidimStrings model.parallelPlotOption.orderedUnitLabels

        yValue1 =
            Maybe.withDefault [] <| List.head yValues

        yValue2 =
            Maybe.withDefault [] <| List.head <| List.drop 1 yValues

        yValue3 =
            Maybe.withDefault [] <| List.head <| List.drop 2 yValues

        yValue4 =
            Maybe.withDefault [] <| List.head <| List.drop 3 yValues

        yValue5 =
            Maybe.withDefault [] <| List.head <| List.drop 4 yValues

        yValue6 =
            Maybe.withDefault [] <| List.head <| List.drop 5 yValues

        yValue7 =
            Maybe.withDefault [] <| List.head <| List.drop 6 yValues

        yValue8 =
            Maybe.withDefault [] <| List.head <| List.drop 7 yValues

        yValue9 =
            Maybe.withDefault [] <| List.head <| List.drop 8 yValues

        yValue10 =
            Maybe.withDefault [] <| List.head <| List.drop 9 yValues

        paths =
            listmap10 pathGenerator yValue1 yValue2 yValue3 yValue4 yValue5 yValue6 yValue7 yValue8 yValue9 yValue10

        pathGenerator : Float -> Float -> Float -> Float -> Float -> Float -> Float -> Float -> Float -> Float -> Path.Path
        pathGenerator y1 y2 y3 y4 y5 y6 y7 y8 y9 y10 =
            let
                yPoints =
                    [ y1, y2, y3, y4, y5, y6, y7, y8, y9, y10 ]

                xPoints =
                    [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 ]

                yPointsScaled =
                    List.map3 (\yS y x -> Just ( Scale.convert xScaleLocal x, Scale.convert yS y + axisPaddig )) yScaleLocals yPoints xPoints
            in
            yPointsScaled |> Shape.line Shape.linearCurve

        lines =
            List.map createLine paths
    in
    svg
        [ viewBox 0 0 (w + 60) (h + 60)
        , TypedSvg.Attributes.width <| TypedSvg.Types.Percent 100
        , TypedSvg.Attributes.height <| TypedSvg.Types.Percent 100
        ]
        ([ style [] [ TypedSvg.Core.text """
            .axis .tick text { display: none; font-size:13px }
            .axis:hover .tick text { display: inline; }
            """ ]
         , g [] []
         ]
            ++ lines
            ++ yaxis
        )


createLine : Path.Path -> Svg msg
createLine path =
    g []
        [ Path.element path
            [ stroke <| Paint <| Color.darkBlue
            , strokeWidth <| Px 0.5
            , fill PaintNone
            ]
        ]


createYAxis : ContinuousScale Float -> Model -> ( Int, ContinuousScale Float ) -> String -> String -> Svg Msg
createYAxis scaleX model scaleY text unit =
    let
        index =
            Tuple.first scaleY

        indexF =
            index |> toFloat

        indexbefore =
            if index == 0 then
                9

            else
                index - 1

        indexafter =
            if index == 9 then
                0

            else
                index + 1

        att1 =
            Maybe.withDefault "Price" <| List.head <| List.drop index model.parallelPlotOption.orderedLabels

        att2 =
            Maybe.withDefault "Price" <| List.head <| List.drop indexafter model.parallelPlotOption.orderedLabels
    in
    g [ transform [ Translate (Scale.convert scaleX indexF) axisPaddig ], class [ "axis" ] ]
        [ g [] [ yAxis (Tuple.second scaleY) ]
        , text_
            [ x -10, y -30, textAnchor AnchorEnd, fontSize <| Px 10, onClick (ChangeOrder index indexbefore) ]
            [ Html.text "<<< " ]
        , text_
            [ x 10, y -30, textAnchor AnchorStart, fontSize <| Px 10, onClick (ChangeOrder index indexafter) ]
            [ Html.text ">>>" ]
        , text_
            [ y -14, textAnchor AnchorMiddle, fontSize <| Px 14, onClick (SetScatterPlotFromParallelPlot att1 att2) ]
            [ Html.text text ]
        , text_
            [ y -5, textAnchor AnchorMiddle, fontSize <| Px 8, onClick (SetScatterPlotFromParallelPlot att1 att2) ]
            [ Html.text unit ]
        ]


yAxis : ContinuousScale Float -> Svg msg
yAxis scale =
    Axis.left [] scale


xScale : List Float -> ContinuousScale Float
xScale values =
    Scale.linear ( padding, w ) (wideExtent values)


yScale : List Float -> ContinuousScale Float
yScale values =
    Scale.linear ( h, 0 ) (wideExtent values)


wideExtent : List Float -> ( Float, Float )
wideExtent values =
    Maybe.withDefault defaultExtent (Statistics.extent values)


filterSPData :
    List Smartphone
    -> Model
    -> List Smartphone
filterSPData smartphone model =
    let
        smartphoneList =
            List.filter (hasIRBlaster model) <|
                List.filter (hasNFC model) <|
                    List.filter (has5G model) <|
                        List.filter (fromBrand model) <|
                            List.filterMap mapToSPData smartphone
    in
    smartphoneList


fromBrand : Model -> Smartphone -> Bool
fromBrand model smartphone =
    if model.parallelPlotOption.selectedFilter.brand == Maybe.withDefault "" smartphone.brand || model.parallelPlotOption.selectedFilter.brand == "all" then
        True

    else
        False


has5G : Model -> Smartphone -> Bool
has5G model smartphone =
    if
        (model.parallelPlotOption.selectedFilter.g5 == "true" && Maybe.withDefault False smartphone.g5)
            || model.parallelPlotOption.selectedFilter.g5
            == "all"
    then
        True

    else if model.parallelPlotOption.selectedFilter.g5 == "false" && not (Maybe.withDefault True smartphone.g5) then
        True

    else
        False


hasNFC : Model -> Smartphone -> Bool
hasNFC model smartphone =
    if
        (model.parallelPlotOption.selectedFilter.nfc == "true" && Maybe.withDefault False smartphone.nfc)
            || model.parallelPlotOption.selectedFilter.nfc
            == "all"
    then
        True

    else if model.parallelPlotOption.selectedFilter.nfc == "false" && not (Maybe.withDefault True smartphone.nfc) then
        True

    else
        False


hasIRBlaster : Model -> Smartphone -> Bool
hasIRBlaster model smartphone =
    if
        (model.parallelPlotOption.selectedFilter.ir_blaster == "true" && Maybe.withDefault False smartphone.ir_blaster)
            || model.parallelPlotOption.selectedFilter.ir_blaster
            == "all"
    then
        True

    else if model.parallelPlotOption.selectedFilter.ir_blaster == "false" && not (Maybe.withDefault True smartphone.ir_blaster) then
        True

    else
        False


mapToSPData : Smartphone -> Maybe Smartphone
mapToSPData smartphone =
    map10
        (\price rating num_cores processor_speed battery fast_charging memory ram screen_size refresh_rate ->
            { id = smartphone.id
            , brand = smartphone.brand
            , model = smartphone.model
            , price = Just price
            , rating = Just rating
            , g5 = smartphone.g5
            , nfc = smartphone.nfc
            , ir_blaster = smartphone.ir_blaster
            , processor_name = smartphone.processor_name
            , processor_brand = smartphone.processor_brand
            , num_cores = Just num_cores
            , processor_speed = Just processor_speed
            , battery = Just battery
            , fast_charging = Just fast_charging
            , memory = Just memory
            , ram = Just ram
            , screen_size = Just screen_size
            , resolution = smartphone.resolution
            , refresh_rate = Just refresh_rate
            , camera = smartphone.camera
            , card = smartphone.card
            , os = smartphone.os
            }
        )
        smartphone.price
        smartphone.rating
        smartphone.num_cores
        smartphone.processor_speed
        smartphone.battery
        smartphone.fast_charging
        smartphone.memory
        smartphone.ram
        smartphone.screen_size
        smartphone.refresh_rate


viewParallelCoordinates : Model -> Html Msg
viewParallelCoordinates model =
    let
        filteredSmartphoneData =
            filterSPData model.data model

        brandList =
            Set.toList <| Set.fromList <| List.filterMap .brand model.data

        brandFilterValues =
            "all" :: brandList

        otherFilterValues =
            [ "all", "true", "false" ]
    in
    Html.div []
        [ Html.div [ class [ "description" ], align "center" ] [ Html.b [] [ Html.text "Hints: By clicking on a axis label, you switch to the scatterplot with clicked axis label as x-value and the following label as y-value. By hovering over an axis, the axis values are displayed. By clicking on the <<< or >>> you can switch the axis backward or forward. You can select different filters to bound the result." ] ]
        , Html.span [ class [ "plotAttributeSelect" ] ] [ b [] [ text " Brand: " ] ]
        , select [ onInput SetBrandParallel ] (List.map (createSelectorOptions model.parallelPlotOption.selectedFilter.brand) brandFilterValues)
        , Html.span [ class [ "plotAttributeSelect" ] ]
            [ b [] [ text " 5G: " ]
            , select [ onInput Set5GParallel ]
                (List.map (createSelectorOptions model.parallelPlotOption.selectedFilter.g5) otherFilterValues)
            ]
        , Html.span [ class [ "plotAttributeSelect" ] ]
            [ b [] [ text " NFC: " ]
            , select [ onInput SetNFCParallel ]
                (List.map (createSelectorOptions model.parallelPlotOption.selectedFilter.nfc) otherFilterValues)
            ]
        , Html.span [ class [ "plotAttributeSelect" ] ]
            [ b [] [ text " IR-Blaster: " ]
            , select [ onInput SetIRBlasterParallel ]
                (List.map (createSelectorOptions model.parallelPlotOption.selectedFilter.ir_blaster) otherFilterValues)
            ]
        , br [] []
        , br [] []
        , parallelCoordinatesPlot model filteredSmartphoneData
        ]


createSelectorOptions : String -> String -> Html Msg
createSelectorOptions select valueOption =
    if select == valueOption then
        option [ value valueOption, selected True ] [ Html.text valueOption ]

    else
        option [ value valueOption ] [ Html.text valueOption ]


type alias MultiDimPoint =
    { pointName : String, value : Float }


type alias MultiDimData =
    { dimDescription : List String
    , data : List (List MultiDimPoint)
    }
